.confidence_intervals {
  font-size: 16px;
}
.confidence {
  padding: 3px;
  display: flex;
}
.level, .label  {
  display: inline-block;
}
.label {
  width: 60px;
}
.confidence_intervals .level {
  font-size: 14px;
  margin-left: 8px;
  margin-right: 8px;
  background-color: #AAA;
  padding: 2px 4px;
  text-align: right;
  font-family: monospace;
  color: white;
  font-weight: bold;
}
.confidence_intervals > * {
  vertical-align: bottom;
}
.flag.flagged {
  background-color: pink !important;
}
.output_class {
  font-weight: bold;
  font-size: 36px;
  flex-grow: 1;
  display: flex;
  align-items: center;
  justify-content: center;
}
