body {
  font-family: 'Open Sans', sans-serif;
  font-size: 12px;
  margin: 0;
}
button, input[type="submit"], input[type="reset"], input[type="text"], input[type="button"], select[type="submit"] {
	background: none;
	color: inherit;
	border: none;
	padding: 0;
	font: inherit;
	cursor: pointer;
	outline: inherit;
  -webkit-appearance: none;
  border-radius: 0;
}
nav, #panels {
  margin-left: 60px;
  margin-right: 60px;
}
nav {
  text-align: center;
  padding: 16px 0 8px;
}
nav img {
  margin-right: auto;
  height: 32px;
}
#panels {
  display: flex;
  flex-flow: row;
  flex-wrap: wrap;
  justify-content: center;
}
