en = {
    "BETA_MESSAGE": "NOTE: Gradio is in beta stage, please report all bugs to: contact.gradio@gmail.com",
    "RUNNING_LOCALLY": "Model is running locally at: {}",
    "NGROK_NO_INTERNET": "Unable to create public link for interface, please check internet connection or try "
                         "restarting python interpreter.",
    "COLAB_NO_LOCAL": "Cannot display local interface on google colab, public link created.",
    "PUBLIC_SHARE_TRUE": "To create a public link, set `share=True` in the argument to `launch()`.",
    "MODEL_PUBLICLY_AVAILABLE_URL": "Model available publicly for 8 hours at: {}"
}
